import numpy as np
import pytest

import autofit as af
from autolens.lens.util import lens_fit_util as util
from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy.util import galaxy_util
from autolens.model.inversion import inversions
from autolens.model.inversion import pixelizations
from autolens.model.inversion import regularization
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.unit.mock.model.mock_profiles import MockLightProfile


class TestLensProfileFit:
    class TestLikelihood:

        def test__1x2_image__no_psf_blurring__tracing_fits_data_with_chi_sq_5(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            ccd_data = ccd.CCDData(5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf,
                                   noise_map=np.ones((3, 4)))
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                            [True, False, False, True],
                                            [True, True, True, True]]), pixel_scale=1.0)

            lens_data_5x5 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(redshift=0.5,
                          light_profile=MockLightProfile(value=1.0, size=2))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_5x5.grid_stack)

            fit = lens_fit.LensProfileFit(lens_data=lens_data_5x5, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (fit.mask_2d == np.array([[True, True, True, True],
                                             [True, False, False, True],
                                             [True, True, True, True]])).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (fit.image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                              [0.0, 5.0, 4.0, 0.0],
                                              [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (fit.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                  [0.0, 1.0, 1.0, 0.0],
                                                  [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.model_image_1d == np.array([1.0, 1.0])).all()
            assert (fit.model_image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 1.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.residual_map_1d == np.array([4.0, 3.0])).all()
            assert (fit.residual_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                     [0.0, 4.0, 3.0, 0.0],
                                                     [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.chi_squared_map_1d == np.array([16.0, 9.0])).all()
            assert (fit.chi_squared_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                        [0.0, 16.0, 9.0, 0.0],
                                                        [0.0, 0.0, 0.0, 0.0]])).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                        25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

        def test__1x2_image__include_psf_blurring__tracing_fits_data_with_chi_sq_4(
                self):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                           [0.0, 1.0, 3.0],
                                           [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            ccd_data = ccd.CCDData(5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf,
                                   noise_map=np.ones((3, 4)))
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                            [True, False, False, True],
                                            [True, True, True, True]]), pixel_scale=1.0)

            lens_data_5x5 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(redshift=0.5,
                          light_profile=MockLightProfile(value=1.0, size=2))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_5x5.grid_stack)

            fit = lens_fit.LensProfileFit(lens_data=lens_data_5x5, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (fit.mask_2d == np.array([[True, True, True, True],
                                             [True, False, False, True],
                                             [True, True, True, True]])).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (fit.image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                              [0.0, 5.0, 4.0, 0.0],
                                              [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (fit.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                  [0.0, 1.0, 1.0, 0.0],
                                                  [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.model_image_1d == np.array([1.0, 4.0])).all()
            assert (fit.model_image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 4.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.residual_map_1d == np.array([4.0, 0.0])).all()
            assert (fit.residual_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                     [0.0, 4.0, 0.0, 0.0],
                                                     [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.chi_squared_map_1d == np.array([16.0, 0.0])).all()
            assert (fit.chi_squared_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                        [0.0, 16.0, 0.0, 0.0],
                                                        [0.0, 0.0, 0.0, 0.0]])).all()

            assert fit.chi_squared == 16.0
            assert fit.reduced_chi_squared == 16.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                        16.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

        def test_hyper_galaxy_changes_noise_above_from_1_to_2__reflected_in_likelihood(
                self):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            # The hyper galaxy increases the noise in both pixels by 1.0, to 2.0.

            # This reduces the chi squared to 2.0 instead of 4.0

            psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                           [0.0, 1.0, 3.0],
                                           [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            ccd_data = ccd.CCDData(5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf,
                                   noise_map=np.ones((3, 4)))
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                            [True, False, False, True],
                                            [True, True, True, True]]), pixel_scale=1.0)

            lens_data_5x5 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(redshift=0.5,
                          light_profile=MockLightProfile(value=1.0, size=2),
                          hyper_galaxy=g.HyperGalaxy(contribution_factor=1.0,
                                                     noise_factor=1.0, noise_power=1.0),
                          hyper_model_image_1d=np.ones(2),
                          hyper_galaxy_image_1d=np.ones(2), hyper_minimum_value=0.0)

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_5x5.grid_stack)

            fit = lens_fit.LensProfileFit(lens_data=lens_data_5x5, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (fit.mask_2d == np.array([[True, True, True, True],
                                             [True, False, False, True],
                                             [True, True, True, True]])).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (fit.image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                              [0.0, 5.0, 4.0, 0.0],
                                              [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.noise_map_1d == np.array([2.0, 2.0])).all()
            assert (fit.noise_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                  [0.0, 2.0, 2.0, 0.0],
                                                  [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.model_image_1d == np.array([1.0, 4.0])).all()
            assert (fit.model_image_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                    [0.0, 1.0, 4.0, 0.0],
                                                    [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.residual_map_1d == np.array([4.0, 0.0])).all()
            assert (fit.residual_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                     [0.0, 4.0, 0.0, 0.0],
                                                     [0.0, 0.0, 0.0, 0.0]])).all()

            assert (fit.chi_squared_map_1d == np.array([4.0, 0.0])).all()
            assert (fit.chi_squared_map_2d == np.array([[0.0, 0.0, 0.0, 0.0],
                                                        [0.0, 4.0, 0.0, 0.0],
                                                        [0.0, 0.0, 0.0, 0.0]])).all()

            assert fit.chi_squared == 4.0
            assert fit.reduced_chi_squared == 4.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.likelihood == -0.5 * (4.0 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0))

    class TestCompareToManual:

        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_5x5):
            g0 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=1.0),
                          mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(redshift=1.0,
                          light_profile=lp.EllipticalSersic(intensity=1.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                         source_galaxies=[g1],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack)

            padded_tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g0], source_galaxies=[g1],
                image_plane_grid_stack=lens_data_5x5.padded_grid_stack)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer,
                                                           padded_tracer=padded_tracer)

            assert lens_data_5x5.noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert lens_data_5x5.noise_map_2d == pytest.approx(fit.noise_map_2d, 1e-4)

            model_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_5x5.convolver_image)
            model_image_2d = lens_data_5x5.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image_2d, 1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=model_image_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=lens_data_5x5.noise_map_1d)

            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=lens_data_5x5.noise_map_2d, mask=lens_data_5x5.mask_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared,
                noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__corresponds_to_blurred_galaxy_images(
                self, lens_data_5x5):
            g0 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=1.0),
                          mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(redshift=1.0,
                          light_profile=lp.EllipticalSersic(intensity=1.0))
            g2 = g.Galaxy(redshift=1.0)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                         source_galaxies=[g1, g2],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            g0_image_plane_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=lens_data_5x5.grid_stack.sub, galaxies=[g0])
            g0_image_plane_bluring_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=lens_data_5x5.grid_stack.blurring, galaxies=[g0])

            g0_blurred_image_plane_image_1d = lens_data_5x5.convolver_image.convolve_image(
                image_array=g0_image_plane_image_1d,
                blurring_array=g0_image_plane_bluring_image_1d)

            g1_image_plane_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=tracer.source_plane.grid_stack.sub, galaxies=[g1])
            g1_image_plane_bluring_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=tracer.source_plane.grid_stack.blurring, galaxies=[g1])

            g1_blurred_image_plane_image_1d = lens_data_5x5.convolver_image.convolve_image(
                image_array=g1_image_plane_image_1d,
                blurring_array=g1_image_plane_bluring_image_1d)

            assert fit.galaxy_image_1d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_1d, 1.0e-4)
            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_1d, 1.0e-4)
            assert (fit.galaxy_image_1d_dict[g2] == np.zeros(9)).all()

            assert fit.model_image_1d == pytest.approx(
                fit.galaxy_image_1d_dict[g0] + fit.galaxy_image_1d_dict[g1], 1.0e-4)

            g0_blurred_image_plane_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=g0_blurred_image_plane_image_1d)
            g1_blurred_image_plane_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=g1_blurred_image_plane_image_1d)

            assert fit.galaxy_image_2d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_2d, 1.0e-4)
            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_2d, 1.0e-4)

            assert fit.model_image_2d == pytest.approx(
                fit.galaxy_image_2d_dict[g0] + fit.galaxy_image_2d_dict[g1], 1.0e-4)

        def test___all_lens_fit_quantities__including_hyper_methods(self,
                                                                    lens_data_5x5):
            g0 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=1.0),
                          mass_profile=mp.SphericalIsothermal(einstein_radius=1.0),
                          hyper_galaxy=g.HyperGalaxy(contribution_factor=1.0,
                                                     noise_factor=1.0, noise_power=1.0),
                          hyper_model_image_1d=np.ones(9),
                          hyper_galaxy_image_1d=np.ones(9), hyper_minimum_value=0.0)
            g1 = g.Galaxy(redshift=1.0,
                          light_profile=lp.EllipticalSersic(intensity=1.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                         source_galaxies=[g1],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack)

            padded_tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g0], source_galaxies=[g1],
                image_plane_grid_stack=lens_data_5x5.padded_grid_stack)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer,
                                                           padded_tracer=padded_tracer)

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_5x5.noise_map_1d)

            hyper_noise_map_1d = lens_data_5x5.noise_map_1d + hyper_noise_1d
            hyper_noise_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=hyper_noise_map_1d)
            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map_2d, 1e-4)

            model_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_5x5.convolver_image)

            model_image_2d = lens_data_5x5.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image_2d, 1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=model_image_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=hyper_noise_map_1d)

            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=hyper_noise_map_2d, mask=lens_data_5x5.mask_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
                self, lens_data_5x5):
            g0 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=1.0),
                          mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(redshift=1.0,
                          light_profile=lp.EllipticalSersic(intensity=1.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                         source_galaxies=[g1],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack)

            padded_tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g0], source_galaxies=[g1],
                image_plane_grid_stack=lens_data_5x5.padded_grid_stack)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer,
                                                           padded_tracer=padded_tracer)

            blurred_profile_image_2d_of_planes = \
                tracer.blurred_profile_image_plane_image_2d_of_planes_from_convolver_image(
                    convolver_image=lens_data_5x5.convolver_image)

            assert (blurred_profile_image_2d_of_planes[0] ==
                    fit.model_image_2d_of_planes[0]).all()
            assert (blurred_profile_image_2d_of_planes[1] ==
                    fit.model_image_2d_of_planes[1]).all()

            unmasked_blurred_profile_image = lens_data_5x5.padded_grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
                psf=lens_data_5x5.psf,
                unmasked_image_1d=padded_tracer.profile_image_plane_image_1d)

            assert (
                        unmasked_blurred_profile_image == fit.unmasked_blurred_image_plane_image).all()

            unmasked_blurred_profile_image_of_planes = \
                padded_tracer.unmasked_blurred_profile_image_plane_image_of_planes_from_psf(
                    psf=lens_data_5x5.psf)

            assert (unmasked_blurred_profile_image_of_planes[0] ==
                    fit.unmasked_blurred_image_plane_image_of_planes[0]).all()
            assert (unmasked_blurred_profile_image_of_planes[1] ==
                    fit.unmasked_blurred_image_plane_image_of_planes[1]).all()

            unmasked_blurred_profile_image_of_galaxies = \
                padded_tracer.unmasked_blurred_profile_image_plane_image_of_plane_and_galaxies_from_psf(
                    psf=lens_data_5x5.psf)

            assert (unmasked_blurred_profile_image_of_galaxies[0][0] ==
                    fit.unmasked_blurred_image_plane_image_of_planes_and_galaxies[0][
                        0]).all()
            assert (unmasked_blurred_profile_image_of_galaxies[1][0] ==
                    fit.unmasked_blurred_image_plane_image_of_planes_and_galaxies[1][
                        0]).all()


class TestLensInversionFit:
    class TestCompareToManual:

        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_5x5):
            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g.Galaxy(redshift=0.5)], source_galaxies=[g0],
                image_plane_grid_stack=lens_data_5x5.grid_stack,
                border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)
            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper, regularization=reg, image_1d=lens_data_5x5.image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix)

            assert inversion.reconstructed_data_1d == pytest.approx(fit.model_image_1d,
                                                                    1e-4)
            assert inversion.reconstructed_data_2d == pytest.approx(fit.model_image_2d,
                                                                    1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=inversion.reconstructed_data_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=lens_data_5x5.noise_map_1d)
            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_5x5.mask_2d, noise_map=lens_data_5x5.noise_map_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared,
                    regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__has_inversion_reconstructed_data(self,
                                                                                lens_data_5x5):
            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(redshift=0.5)
            g1 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0],
                                                         source_galaxies=[g1],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack,
                                                         border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)
            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper, regularization=reg, image_1d=lens_data_5x5.image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix)

            assert (fit.galaxy_image_1d_dict[g0] == np.zeros(9)).all()

            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                inversion.reconstructed_data_1d, 1.0e-4)

            assert fit.model_image_1d == pytest.approx(fit.galaxy_image_1d_dict[g1],
                                                       1.0e-4)

            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                inversion.reconstructed_data_2d, 1.0e-4)

            assert fit.model_image_2d == pytest.approx(fit.galaxy_image_2d_dict[g1],
                                                       1.0e-4)

        def test___all_lens_fit_quantities__include_hyper_methods(self, lens_data_5x5):
            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg,
                          hyper_galaxy=g.HyperGalaxy(contribution_factor=1.0,
                                                     noise_factor=1.0, noise_power=1.0),
                          hyper_model_image_1d=np.ones(9),
                          hyper_galaxy_image_1d=np.ones(9), hyper_minimum_value=0.0)

            tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g.Galaxy(redshift=0.5)], source_galaxies=[g0],
                image_plane_grid_stack=lens_data_5x5.grid_stack,
                border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_5x5.noise_map_1d)
            hyper_noise_map_1d = lens_data_5x5.noise_map_1d + hyper_noise_1d
            hyper_noise_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=hyper_noise_map_1d)
            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map_2d, 1e-4)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)
            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper, regularization=reg, image_1d=lens_data_5x5.image_1d,
                noise_map_1d=hyper_noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix)

            assert inversion.reconstructed_data_1d == pytest.approx(fit.model_image_1d,
                                                                    1e-4)
            assert inversion.reconstructed_data_2d == pytest.approx(fit.model_image_2d,
                                                                    1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=inversion.reconstructed_data_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=hyper_noise_map_1d)
            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_5x5.mask_2d, noise_map=hyper_noise_map_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared,
                    regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
                self, lens_data_5x5):
            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=[g.Galaxy(redshift=0.5)], source_galaxies=[g0],
                image_plane_grid_stack=lens_data_5x5.grid_stack,
                border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper, regularization=reg, image_1d=lens_data_5x5.image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix)

            assert None == fit.model_image_2d_of_planes[0]
            assert inversion.reconstructed_data_2d == pytest.approx(
                fit.model_image_2d_of_planes[1], 1.0e-4)

            with pytest.raises(Exception):
                fit.unmasked_blurred_image_plane_image
                fit.unmasked_blurred_image_plane_image_of_planes
                fit.unmasked_blurred_image_plane_image_of_planes_and_galaxies


class TestLensProfileInversionFit:
    class TestCompareToManual:

        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_5x5):
            galaxy_light = g.Galaxy(redshift=0.5,
                                    light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light],
                                                         source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack,
                                                         border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_5x5.convolver_image)

            blurred_profile_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=blurred_profile_image_1d)

            assert blurred_profile_image_1d == pytest.approx(
                fit.blurred_profile_image_1d, 1e-4)
            assert blurred_profile_image_2d == pytest.approx(
                fit.blurred_profile_image_2d, 1e-4)

            profile_subtracted_image_1d = lens_data_5x5.image_1d - blurred_profile_image_1d
            profile_subtracted_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=profile_subtracted_image_1d)

            assert profile_subtracted_image_1d == pytest.approx(
                fit.profile_subtracted_image_1d)
            assert profile_subtracted_image_2d == pytest.approx(
                fit.profile_subtracted_image_2d)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix, mapper=mapper,
                regularization=reg)

            model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
            model_image_2d = lens_data_5x5.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image_2d, 1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=model_image_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=lens_data_5x5.noise_map_1d)
            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_5x5.mask_2d, noise_map=lens_data_5x5.noise_map_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared,
                    regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__has_blurred_profile_images_and_inversion_reconstructed_data(
                self, lens_data_5x5):
            g0 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(redshift=0.5,
                          light_profile=lp.EllipticalSersic(intensity=2.0))
            g2 = g.Galaxy(redshift=0.5)

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1, g2],
                                                         source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack,
                                                         border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            g0_image_plane_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=lens_data_5x5.grid_stack.sub, galaxies=[g0])
            g0_image_plane_bluring_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=lens_data_5x5.grid_stack.blurring, galaxies=[g0])

            g0_blurred_image_plane_image_1d = lens_data_5x5.convolver_image.convolve_image(
                image_array=g0_image_plane_image_1d,
                blurring_array=g0_image_plane_bluring_image_1d)

            g1_image_plane_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=tracer.source_plane.grid_stack.sub, galaxies=[g1])
            g1_image_plane_bluring_image_1d = galaxy_util.intensities_of_galaxies_from_grid(
                grid=tracer.source_plane.grid_stack.blurring, galaxies=[g1])

            g1_blurred_image_plane_image_1d = lens_data_5x5.convolver_image.convolve_image(
                image_array=g1_image_plane_image_1d,
                blurring_array=g1_image_plane_bluring_image_1d)

            blurred_profile_image_1d = g0_blurred_image_plane_image_1d + g1_blurred_image_plane_image_1d

            profile_subtracted_image_1d = lens_data_5x5.image_1d - blurred_profile_image_1d
            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix, mapper=mapper,
                regularization=reg)

            assert fit.galaxy_image_1d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_1d, 1.0e-4)
            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_1d, 1.0e-4)
            assert (fit.galaxy_image_1d_dict[g2] == np.zeros(9)).all()
            assert fit.galaxy_image_1d_dict[galaxy_pix] == pytest.approx(
                inversion.reconstructed_data_1d, 1.0e-4)

            assert fit.model_image_1d == pytest.approx(
                fit.galaxy_image_1d_dict[g0] + fit.galaxy_image_1d_dict[g1] +
                fit.galaxy_image_1d_dict[galaxy_pix], 1.0e-4)

            g0_blurred_image_plane_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=g0_blurred_image_plane_image_1d)
            g1_blurred_image_plane_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=g1_blurred_image_plane_image_1d)

            assert fit.galaxy_image_2d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_2d, 1.0e-4)
            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_2d, 1.0e-4)
            assert fit.galaxy_image_2d_dict[galaxy_pix] == pytest.approx(
                inversion.reconstructed_data_2d, 1.0e-4)

            assert fit.model_image_2d == pytest.approx(fit.galaxy_image_2d_dict[g0] +
                                                       fit.galaxy_image_2d_dict[g1] +
                                                       inversion.reconstructed_data_2d,
                                                       1.0e-4)

        def test___all_lens_fit_quantities__include_hyper_methods(self, lens_data_5x5):
            galaxy_light = g.Galaxy(redshift=0.5,
                                    light_profile=lp.EllipticalSersic(intensity=1.0),
                                    hyper_galaxy=g.HyperGalaxy(contribution_factor=1.0,
                                                               noise_factor=1.0,
                                                               noise_power=1.0),
                                    hyper_model_image_1d=np.ones(9),
                                    hyper_galaxy_image_1d=np.ones(9),
                                    hyper_minimum_value=0.0)

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light],
                                                         source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack,
                                                         border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_5x5.noise_map_1d)
            hyper_noise_map_1d = lens_data_5x5.noise_map_1d + hyper_noise_1d
            hyper_noise_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=hyper_noise_map_1d)
            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map_2d, 1e-4)

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_5x5.convolver_image)

            blurred_profile_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=blurred_profile_image_1d)

            assert blurred_profile_image_1d == pytest.approx(
                fit.blurred_profile_image_1d, 1e-4)
            assert blurred_profile_image_2d == pytest.approx(
                fit.blurred_profile_image_2d, 1e-4)

            profile_subtracted_image_1d = lens_data_5x5.image_1d - blurred_profile_image_1d
            profile_subtracted_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=profile_subtracted_image_1d)

            assert profile_subtracted_image_1d == pytest.approx(
                fit.profile_subtracted_image_1d)
            assert profile_subtracted_image_2d == pytest.approx(
                fit.profile_subtracted_image_2d)

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d, noise_map_1d=hyper_noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix, mapper=mapper,
                regularization=reg)

            model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
            model_image_2d = lens_data_5x5.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image_2d, 1e-4)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_5x5.image_1d, mask=lens_data_5x5.mask_1d,
                model_data=model_image_1d)
            residual_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=residual_map_1d)

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map_2d, 1e-4)

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d, mask=lens_data_5x5.mask_1d,
                noise_map=hyper_noise_map_1d)
            chi_squared_map_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=chi_squared_map_1d)

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(fit.chi_squared_map_2d, 1e-4)

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_5x5.mask_2d)

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_5x5.mask_2d, noise_map=hyper_noise_map_2d)

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared,
                    regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
                self, lens_data_5x5):
            galaxy_light = g.Galaxy(redshift=0.5,
                                    light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light],
                                                         source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_5x5.grid_stack,
                                                         border=None)

            fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                           tracer=tracer)

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_5x5.convolver_image)

            blurred_profile_image_2d = lens_data_5x5.map_to_scaled_array(
                array_1d=blurred_profile_image_1d)

            profile_subtracted_image_1d = lens_data_5x5.image_1d - blurred_profile_image_1d

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_5x5.grid_stack, border=None)

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_5x5.noise_map_1d,
                convolver=lens_data_5x5.convolver_mapping_matrix, mapper=mapper,
                regularization=reg)

            assert blurred_profile_image_2d == pytest.approx(
                fit.model_image_2d_of_planes[0], 1.0e-4)
            assert inversion.reconstructed_data_2d == pytest.approx(
                fit.model_image_2d_of_planes[1], 1.0e-4)


class MockTracerPositions:

    def __init__(self, positions, noise=None):
        self.positions = positions
        self.noise = noise


class TestPositionFit:

    def test__x1_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(2)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(1.0) + np.square(3.0))

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [-9.0, -4.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_positions__mock_position_tracer__maximum_separation_is_correct(
            self):
        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[np.array([[-2.0, -4.0], [1.0, 3.0], [0.1, 0.1], [-0.1, -0.1],
                                 [0.3, 0.4], [-0.6, 0.5]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(
            positions=[np.array([[8.0, 4.0], [8.0, 4.0], [-9.0, -4.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_sets_of_positions__multiple_sets_of_max_distances(self):
        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                       np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                       np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separations[0] == 1.0
        assert fit.maximum_separations[1] == np.sqrt(18)
        assert fit.maximum_separations[2] == np.sqrt(18)

    def test__likelihood__is_sum_of_separations_divided_by_noise(self):
        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                       np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                       np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.chi_squared_map[0] == 1.0
        assert fit.chi_squared_map[1] == pytest.approx(18.0, 1e-4)
        assert fit.chi_squared_map[2] == pytest.approx(18.0, 1e-4)
        assert fit.figure_of_merit == pytest.approx(-0.5 * (1.0 + 18 + 18), 1e-4)

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=2.0)
        assert fit.chi_squared_map[0] == (1.0 / 2.0) ** 2.0
        assert fit.chi_squared_map[1] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0,
                                                       1e-4)
        assert fit.chi_squared_map[2] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0,
                                                       1e-4)
        assert fit.figure_of_merit == pytest.approx(
            -0.5 * ((1.0 / 2.0) ** 2.0 + (np.sqrt(18.0) / 2.0) ** 2.0 +
                    (np.sqrt(18.0) / 2.0) ** 2.0), 1e-4)

    def test__threshold__if_not_met_returns_ray_tracing_exception(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separation_within_threshold(threshold=100.0)
        assert not fit.maximum_separation_within_threshold(threshold=0.1)



import numpy as np
from astropy import cosmology as cosmo

import autofit as af
from autolens import exc, dimensions as dim
from autolens.data.array import grids
from autolens.data.array import scaled_array
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.model.galaxy.util import galaxy_util


class AbstractPlane(object):

    def __init__(self, redshift, galaxies=None, cosmology=cosmo.Planck15):
        """An abstract plane which represents a set of galaxies that are close to one another in redshift-space.

        From an abstract plane, cosmological quantities like angular diameter distances can be computed. If the  \
        redshift of the plane is input as *None*, quantities that depend on a redshift are returned as *None*.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of galaxies in this plane.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        self.redshift = redshift
        self.galaxies = galaxies
        self.cosmology = cosmology

    @property
    def galaxy_redshifts(self):
        return [galaxy.redshift for galaxy in self.galaxies]

    @property
    def arcsec_per_kpc(self):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=self.redshift,
            cosmology=self.cosmology)

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    def angular_diameter_distance_to_earth_in_units(self, unit_length='arcsec'):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology, unit_length=unit_length)

    def cosmic_average_density_in_units(self, unit_length='arcsec',
                                        unit_mass='solMass'):
        return cosmology_util.cosmic_average_density_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology, unit_length=unit_length,
            unit_mass=unit_mass)

    @property
    def has_light_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_light_profile, self.galaxies)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_mass_profile, self.galaxies)))

    @property
    def has_pixelization(self):
        return any(list(map(lambda galaxy: galaxy.has_pixelization, self.galaxies)))

    @property
    def has_regularization(self):
        return any(list(map(lambda galaxy: galaxy.has_regularization, self.galaxies)))

    @property
    def regularization(self):

        galaxies_with_regularization = list(
            filter(lambda galaxy: galaxy.has_regularization, self.galaxies))

        if len(galaxies_with_regularization) == 0:
            return None
        if len(galaxies_with_regularization) == 1:
            return galaxies_with_regularization[0].regularization
        elif len(galaxies_with_regularization) > 1:
            raise exc.PixelizationException(
                'The number of galaxies with regularizations in one plane is above 1')

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda galaxy: galaxy.has_hyper_galaxy, self.galaxies)))

    @property
    def contribution_maps_1d_of_galaxies(self):

        contribution_maps_1d = []

        for galaxy in self.galaxies:

            if galaxy.hyper_galaxy is not None:

                contribution_map = galaxy.hyper_galaxy.contribution_map_from_hyper_images(
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d)

                contribution_maps_1d.append(contribution_map)

            else:

                contribution_maps_1d.append(None)

        return contribution_maps_1d

    @property
    def centres_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [galaxy for galaxy in self.galaxies if
                                       galaxy.has_mass_profile]

        mass_profile_centres = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_centres[galaxy_index] = [profile.centre for profile in
                                                  galaxy.mass_profiles]
        return mass_profile_centres

    @property
    def axis_ratios_of_galaxy_mass_profiles(self):
        galaxies_with_mass_profiles = [galaxy for galaxy in self.galaxies if
                                       galaxy.has_mass_profile]

        mass_profile_axis_ratios = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_axis_ratios[galaxy_index] = [profile.axis_ratio for profile in
                                                      galaxy.mass_profiles]
        return mass_profile_axis_ratios

    @property
    def phis_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [galaxy for galaxy in self.galaxies if
                                       galaxy.has_mass_profile]

        mass_profile_phis = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_phis[galaxy_index] = [profile.phi for profile in
                                               galaxy.mass_profiles]
        return mass_profile_phis

    def luminosities_of_galaxies_within_circles_in_units(self, radius: dim.Length,
                                                         unit_luminosity='eps',
                                                         exposure_time=None):
        """Compute the total luminosity of all galaxies in this plane within a circle of specified radius.

        See *galaxy.light_within_circle* and *light_profiles.light_within_circle* for details \
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(map(lambda galaxy: galaxy.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity=unit_luminosity, exposure_time=exposure_time,
            cosmology=self.cosmology),
                        self.galaxies))

    def luminosities_of_galaxies_within_ellipses_in_units(self, major_axis: dim.Length,
                                                          unit_luminosity='eps',
                                                          exposure_time=None):
        """
        Compute the total luminosity of all galaxies in this plane within a ellipse of specified major-axis.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the photometric zeropoint).

        See *galaxy.light_within_ellipse* and *light_profiles.light_within_ellipse* for details
        of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(map(lambda galaxy: galaxy.luminosity_within_ellipse_in_units(
            major_axis=major_axis, unit_luminosity=unit_luminosity,
            exposure_time=exposure_time,
            cosmology=self.cosmology),
                        self.galaxies))

    def masses_of_galaxies_within_circles_in_units(self, radius: dim.Length,
                                                   unit_mass='solMass',
                                                   redshift_source=None):
        """Compute the total mass of all galaxies in this plane within a circle of specified radius.

        See *galaxy.angular_mass_within_circle* and *mass_profiles.angular_mass_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        redshift_source
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The units the mass is returned in (angular | solMass).

        """
        return list(map(lambda galaxy: galaxy.mass_within_circle_in_units(
            radius=radius, unit_mass=unit_mass, redshift_source=redshift_source,
            cosmology=self.cosmology),
                        self.galaxies))

    def masses_of_galaxies_within_ellipses_in_units(self, major_axis: dim.Length,
                                                    unit_mass='solMass',
                                                    redshift_source=None):
        """Compute the total mass of all galaxies in this plane within a ellipse of specified major-axis.

        See *galaxy.angular_mass_within_ellipse* and *mass_profiles.angular_mass_within_ellipse* for details \
        of how this is performed.

        Parameters
        ----------
        redshift_source
        unit_mass
        major_axis : float
            The major-axis radius of the ellipse.

        """
        return list(map(lambda galaxy: galaxy.mass_within_ellipse_in_units(
            major_axis=major_axis, unit_mass=unit_mass, redshift_source=redshift_source,
            cosmology=self.cosmology),
                        self.galaxies))

    def einstein_radius_in_units(self, unit_length='arcsec'):

        if self.has_mass_profile:
            return sum(filter(None,
                              list(map(lambda galaxy: galaxy.einstein_radius_in_units(
                                  unit_length=unit_length, cosmology=self.cosmology),
                                       self.galaxies))))

    def einstein_mass_in_units(self, unit_mass='solMass', redshift_source=None):

        if self.has_mass_profile:
            return sum(filter(None,
                              list(map(lambda galaxy: galaxy.einstein_mass_in_units(
                                  unit_mass=unit_mass, redshift_source=redshift_source,
                                  cosmology=self.cosmology),
                                       self.galaxies))))

    # noinspection PyUnusedLocal
    def summarize_in_units(self, radii, whitespace=80,
                           unit_length='arcsec', unit_luminosity='eps',
                           unit_mass='solMass',
                           redshift_source=None, **kwargs):

        summary = ['Plane\n']
        prefix_plane = ''

        summary += [af.text_util.label_and_value_string(
            label=prefix_plane + 'redshift', value=self.redshift, whitespace=whitespace,
            format_string='{:.2f}')]

        summary += [af.text_util.label_and_value_string(
            label=prefix_plane + 'kpc_per_arcsec', value=self.kpc_per_arcsec,
            whitespace=whitespace,
            format_string='{:.2f}')]

        angular_diameter_distance_to_earth = self.angular_diameter_distance_to_earth_in_units(
            unit_length=unit_length)

        summary += [af.text_util.label_and_value_string(
            label=prefix_plane + 'angular_diameter_distance_to_earth',
            value=angular_diameter_distance_to_earth,
            whitespace=whitespace,
            format_string='{:.2f}')]

        for galaxy in self.galaxies:
            summary += ['\n']
            summary += galaxy.summarize_in_units(radii=radii, whitespace=whitespace,
                                                 unit_length=unit_length,
                                                 unit_luminosity=unit_luminosity,
                                                 unit_mass=unit_mass,
                                                 redshift_source=redshift_source,
                                                 cosmology=self.cosmology)

        return summary


class AbstractGriddedPlane(AbstractPlane):

    def __init__(self, redshift, grid_stack, border, compute_deflections, galaxies=None,
                 cosmology=cosmo.Planck15):
        """An abstract plane which represents a set of galaxies that are close to one another in redshift-space and \
        have an associated grid on which lensing calcuations are performed.

        From an abstract plane grid, the surface-density, potential and deflection angles of the galaxies can be \
        computed.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        grid_stack : masks.GridStack
            The stack of grid_stacks of (y,x) arc-second coordinates of this plane.
        border : masks.RegularGridBorder
            The borders of the regular-grid, which is used to relocate demagnified traced regular-pixel to the \
            source-plane borders.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        super(AbstractGriddedPlane, self).__init__(redshift=redshift, galaxies=galaxies,
                                                   cosmology=cosmology)

        self.grid_stack = grid_stack
        self.border = border

        if compute_deflections:

            def calculate_deflections(grid):

                if galaxies:
                    return sum(map(lambda galaxy: galaxy.deflections_from_grid(grid),
                                   galaxies))
                else:
                    return np.full((grid.shape[0], 2), 0.0)

            self.deflections_stack = grid_stack.apply_function(calculate_deflections)

        else:

            self.deflections_stack = None

    def trace_grid_stack_to_next_plane(self):
        """Trace this plane's grid_stacks to the next plane, using its deflection angles."""

        def minus(grid, deflections):
            return grid - deflections

        return self.grid_stack.map_function(minus, self.deflections_stack)

    @property
    def profile_image_plane_image_2d(self):
        return self.grid_stack.scaled_array_2d_from_array_1d(
            self.profile_image_plane_image_1d)

    @property
    def profile_image_plane_image_2d_for_simulation(self):
        if not self.has_padded_grid_stack:
            raise exc.RayTracingException(
                'To retrieve an image plane image for a simulation, the grid stack in the plane'
                'must be a padded grid stack')
        return self.grid_stack.regular.padded_array_2d_from_padded_array_1d(
            padded_array_1d=self.profile_image_plane_image_1d)

    @property
    def profile_image_plane_image_1d(self):
        return galaxy_util.intensities_of_galaxies_from_grid(grid=self.grid_stack.sub,
                                                             galaxies=self.galaxies)

    @property
    def profile_image_plane_image_1d_of_galaxies(self):
        return list(map(self.profile_image_plane_image_1d_of_galaxy, self.galaxies))

    def profile_image_plane_image_1d_of_galaxy(self, galaxy):
        return galaxy_util.intensities_of_galaxies_from_grid(grid=self.grid_stack.sub,
                                                             galaxies=[galaxy])

    @property
    def profile_image_plane_blurring_image_1d(self):
        return galaxy_util.intensities_of_galaxies_from_grid(
            grid=self.grid_stack.blurring, galaxies=self.galaxies)

    @property
    def profile_image_plane_blurring_image_1d_of_galaxies(self):
        return list(
            map(self.profile_image_plane_blurring_image_1d_of_galaxy, self.galaxies))

    def profile_image_plane_blurring_image_1d_of_galaxy(self, galaxy):
        return galaxy_util.intensities_of_galaxies_from_grid(
            grid=self.grid_stack.blurring, galaxies=[galaxy])

    @property
    def convergence(self):
        convergence_1d = galaxy_util.convergence_of_galaxies_from_grid(
            grid=self.grid_stack.sub.unlensed_sub_grid, galaxies=self.galaxies)
        return self.grid_stack.scaled_array_2d_from_array_1d(array_1d=convergence_1d)

    @property
    def potential(self):
        potential_1d = galaxy_util.potential_of_galaxies_from_grid(
            grid=self.grid_stack.sub.unlensed_sub_grid,
            galaxies=self.galaxies)
        return self.grid_stack.scaled_array_2d_from_array_1d(array_1d=potential_1d)

    @property
    def deflections_y(self):
        return self.grid_stack.scaled_array_2d_from_array_1d(self.deflections_1d[:, 0])

    @property
    def deflections_x(self):
        return self.grid_stack.scaled_array_2d_from_array_1d(self.deflections_1d[:, 1])

    @property
    def deflections_1d(self):
        return galaxy_util.deflections_of_galaxies_from_grid(
            grid=self.grid_stack.sub.unlensed_sub_grid,
            galaxies=self.galaxies)

    @property
    def has_padded_grid_stack(self):
        return issubclass(type(self.grid_stack.regular), grids.PaddedRegularGrid)

    @property
    def plane_image(self):
        return lens_util.plane_image_of_galaxies_from_grid(
            shape=self.grid_stack.regular.mask.shape,
            grid=self.grid_stack.regular,
            galaxies=self.galaxies)

    @property
    def mapper(self):

        galaxies_with_pixelization = list(
            filter(lambda galaxy: galaxy.has_pixelization, self.galaxies))

        if len(galaxies_with_pixelization) == 0:
            return None
        if len(galaxies_with_pixelization) == 1:

            pixelization = galaxies_with_pixelization[0].pixelization

            return pixelization.mapper_from_grid_stack_and_border(
                grid_stack=self.grid_stack, border=self.border,
                hyper_image=galaxies_with_pixelization[0].hyper_galaxy_image_1d)

        elif len(galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                'The number of galaxies with pixelizations in one plane is above 1')

    @property
    def yticks(self):
        """Compute the yticks labels of this grid_stack, used for plotting the y-axis ticks when visualizing an image \
        """
        return np.linspace(np.amin(self.grid_stack.regular[:, 0]),
                           np.amax(self.grid_stack.regular[:, 0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid_stack, used for plotting the x-axis ticks when visualizing an \
        image"""
        return np.linspace(np.amin(self.grid_stack.regular[:, 1]),
                           np.amax(self.grid_stack.regular[:, 1]), 4)


class AbstractDataPlane(AbstractGriddedPlane):

    def blurred_profile_image_plane_image_1d_from_convolver_image(self,
                                                                  convolver_image):
        return convolver_image.convolve_image(
            image_array=self.profile_image_plane_image_1d,
            blurring_array=self.profile_image_plane_blurring_image_1d)

    def blurred_profile_image_plane_images_1d_of_galaxies_from_convolver_image(self,
                                                                               convolver_image):

        return list(map(
            lambda profile_image_plane_image_1d, profile_image_plane_blurring_image_1d:
            convolver_image.convolve_image(image_array=profile_image_plane_image_1d,
                                           blurring_array=profile_image_plane_blurring_image_1d),
            self.profile_image_plane_image_1d_of_galaxies,
            self.profile_image_plane_blurring_image_1d_of_galaxies))

    def unmasked_blurred_profile_image_plane_image_from_psf(self, psf):

        if not self.has_padded_grid_stack:
            raise exc.RayTracingException(
                'To retrieve an unmasked image, the grid stack of a plane tracer'
                'must be a padded grid stack')

        return self.grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
            psf=psf, unmasked_image_1d=self.profile_image_plane_image_1d)

    def unmasked_blurred_profile_image_plane_images_of_galaxies_from_psf(self, psf):

        if not self.has_padded_grid_stack:
            raise exc.RayTracingException(
                'To retrieve an unmasked image, the grid stack of a plane tracer'
                'must be a padded grid stack')

        if self.has_pixelization:
            raise exc.RayTracingException(
                'As unmasked blurred image plane image cannot be returned frmm a plane'
                'with a pixelization')

        return list(map(lambda image_plane_image_1d:
                        self.grid_stack.unmasked_blurred_image_from_psf_and_unmasked_image(
                            psf=psf, unmasked_image_1d=image_plane_image_1d),
                        self.profile_image_plane_image_1d_of_galaxies))

    def hyper_noise_map_1d_from_noise_map_1d(self, noise_map_1d):
        hyper_noise_maps_1d = self.hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(
            noise_map_1d=noise_map_1d)
        hyper_noise_maps_1d = [hyper_noise_map for hyper_noise_map in
                               hyper_noise_maps_1d if
                               hyper_noise_map is not None]
        return sum(hyper_noise_maps_1d)

    def hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(self, noise_map_1d):
        """For a contribution map and noise-map, use the model hyper galaxies to compute a scaled noise-map.

        Parameters
        -----------
        noise_map_1d : ccd.NoiseMap or ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        """
        hyper_noise_maps_1d = []

        for galaxy in self.galaxies:
            if galaxy.hyper_galaxy is not None:

                hyper_noise_map_1d = galaxy.hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    noise_map=noise_map_1d,
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d)

                hyper_noise_maps_1d.append(hyper_noise_map_1d)

            else:

                hyper_noise_maps_1d.append(None)

        return hyper_noise_maps_1d


class Plane(AbstractDataPlane):

    def __init__(self, galaxies, grid_stack, redshift=None, border=None,
                 compute_deflections=True,
                 cosmology=cosmo.Planck15):
        """A plane of galaxies where all galaxies are at the same redshift.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        grid_stack : masks.GridStack
            The stack of grid_stacks of (y,x) arc-second coordinates of this plane.
        border : masks.RegularGridBorder
            The borders of the regular-grid, which is used to relocate demagnified traced regular-pixel to the \
            source-plane borders.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        if redshift is None:

            if not galaxies:
                raise exc.RayTracingException(
                    'A redshift and no galaxies were input to a Plane. A redshift for the Plane therefore cannot be'
                    'determined')
            elif not all(
                    [galaxies[0].redshift == galaxy.redshift for galaxy in galaxies]):
                raise exc.RayTracingException(
                    'A redshift and two or more galaxies with different redshifts were input to a Plane. A unique '
                    'Redshift for the Plane therefore cannot be determined')
            else:
                redshift = galaxies[0].redshift

        super(Plane, self).__init__(redshift=redshift, galaxies=galaxies,
                                    grid_stack=grid_stack,
                                    border=border,
                                    compute_deflections=compute_deflections,
                                    cosmology=cosmology)


class PlanePositions(object):

    def __init__(self, redshift, galaxies, positions, compute_deflections=True,
                 cosmology=None):
        """A plane represents a set of galaxies at a given redshift in a ray-tracer_normal and the positions of image-plane \
        coordinates which mappers close to one another in the source-plane.

        Parameters
        -----------
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        positions : [[[]]]
            The (y,x) arc-second coordinates of image-plane pixels which (are expected to) mappers to the same
            location(s) in the final source-plane.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        """

        self.redshift = redshift
        self.galaxies = galaxies
        self.positions = positions

        if compute_deflections:
            def calculate_deflections(pos):
                return sum(
                    map(lambda galaxy: galaxy.deflections_from_grid(pos), galaxies))

            self.deflections = list(
                map(lambda pos: calculate_deflections(pos), self.positions))

        self.cosmology = cosmology

    def trace_to_next_plane(self):
        """Trace the positions to the next plane."""
        return list(
            map(lambda positions, deflections: np.subtract(positions, deflections),
                self.positions, self.deflections))


class PlaneImage(scaled_array.ScaledRectangularPixelArray):

    def __init__(self, array, pixel_scales, grid, origin=(0.0, 0.0)):
        self.grid = grid
        super(PlaneImage, self).__init__(array=array, pixel_scales=pixel_scales,
                                         origin=origin)



import numpy as np
import pytest
from astropy import cosmology as cosmo

import autofit as af
from autolens import exc
from autolens.lens import lens_fit
from autolens.lens import ray_tracing as rt
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.inversion import pixelizations as px
from autolens.model.inversion import regularization as rg
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.pipeline.phase import phase_hyper
from autolens.pipeline.phase import phase_imaging
from test.unit.mock.pipeline import mock_pipeline


@pytest.fixture(name="lens_galaxy")
def make_lens_galaxy():
    return g.Galaxy(redshift=1.0, light=lp.SphericalSersic(),
                    mass=mp.SphericalIsothermal())


@pytest.fixture(name="source_galaxy")
def make_source_galaxy():
    return g.Galaxy(redshift=2.0, light=lp.SphericalSersic())


@pytest.fixture(name="lens_galaxies")
def make_lens_galaxies(lens_galaxy):
    lens_galaxies = af.ModelInstance()
    lens_galaxies.lens = lens_galaxy
    return lens_galaxies


@pytest.fixture(name="all_galaxies")
def make_all_galaxies(lens_galaxy, source_galaxy):
    galaxies = af.ModelInstance()
    galaxies.lens = lens_galaxy
    galaxies.source = source_galaxy
    return galaxies


@pytest.fixture(name="lens_instance")
def make_lens_instance(lens_galaxies):
    instance = af.ModelInstance()
    instance.lens_galaxies = lens_galaxies
    return instance


@pytest.fixture(name="lens_result")
def make_lens_result(lens_data_5x5, lens_instance):
    return phase_imaging.LensPlanePhase.Result(
        constant=lens_instance, figure_of_merit=1.0, previous_variable=af.ModelMapper(),
        gaussian_tuples=None,
        analysis=phase_imaging.LensPlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=cosmo.Planck15, positions_threshold=1.0),
        optimizer=None)


@pytest.fixture(name="lens_source_instance")
def make_lens_source_instance(lens_galaxy, source_galaxy):
    source_galaxies = af.ModelInstance()
    lens_galaxies = af.ModelInstance()
    source_galaxies.source = source_galaxy
    lens_galaxies.lens = lens_galaxy

    instance = af.ModelInstance()
    instance.source_galaxies = source_galaxies
    instance.lens_galaxies = lens_galaxies
    return instance


@pytest.fixture(name="lens_source_result")
def make_lens_source_result(lens_data_5x5, lens_source_instance):
    return phase_imaging.LensSourcePlanePhase.Result(
        constant=lens_source_instance, figure_of_merit=1.0,
        previous_variable=af.ModelMapper(), gaussian_tuples=None,
        analysis=phase_imaging.LensSourcePlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=cosmo.Planck15, positions_threshold=1.0),
        optimizer=None)


@pytest.fixture(name="multi_plane_instance")
def make_multi_plane_instance(all_galaxies):
    instance = af.ModelInstance()
    instance.galaxies = all_galaxies
    return instance


@pytest.fixture(name="multi_plane_result")
def make_multi_plane_result(lens_data_5x5, multi_plane_instance):
    return phase_imaging.MultiPlanePhase.Result(
        constant=multi_plane_instance, figure_of_merit=1.0,
        previous_variable=af.ModelMapper(), gaussian_tuples=None,
        analysis=phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=cosmo.Planck15, positions_threshold=1.0),
        optimizer=None)


class TestPixelization(object):

    def test_make_pixelization_variable(self):
        instance = af.ModelInstance()
        mapper = af.ModelMapper()

        mapper.lens_galaxy = gm.GalaxyModel(redshift=g.Redshift,
                                            pixelization=px.Rectangular,
                                            regularization=rg.Constant)
        mapper.source_galaxy = gm.GalaxyModel(
            redshift=g.Redshift,
            light=lp.EllipticalLightProfile)

        assert mapper.prior_count == 9

        instance.lens_galaxy = g.Galaxy(pixelization=px.Rectangular(),
                                        regularization=rg.Constant(), redshift=1.0)
        instance.source_galaxy = g.Galaxy(redshift=1.0,
                                          light=lp.EllipticalLightProfile())

        phase_hyper.HyperPixelizationPhase.transfer_classes(instance, mapper)

        assert mapper.prior_count == 3
        assert mapper.lens_galaxy.redshift == 1.0
        assert mapper.source_galaxy.light.axis_ratio == 1.0


class TestImagePassing(object):

    def test_path_galaxy_tuples(self, lens_result, lens_galaxy):
        assert lens_result.path_galaxy_tuples == [
            (("lens_galaxies", "lens"), lens_galaxy)]

    def test_lens_source_galaxy_dict(self, lens_source_result, lens_galaxy,
                                     source_galaxy):
        assert lens_source_result.path_galaxy_tuples == [
            (("source_galaxies", "source"), source_galaxy),
            (("lens_galaxies", "lens"), lens_galaxy)
        ]

    def test_multi_plane_galaxy_dict(self, multi_plane_result, lens_galaxy,
                                     source_galaxy):
        assert multi_plane_result.path_galaxy_tuples == [
            (("galaxies", "lens"), lens_galaxy),
            (("galaxies", "source"), source_galaxy)
        ]

    def test_lens_image_dict(self, lens_result):
        image_dict = lens_result.image_2d_dict
        assert isinstance(image_dict[("lens_galaxies", "lens")], np.ndarray)

    def test_lens_source_image_dict(self, lens_source_result):
        image_dict = lens_source_result.image_2d_dict
        assert isinstance(image_dict[("lens_galaxies", "lens")], np.ndarray)
        assert isinstance(image_dict[("source_galaxies", "source")], np.ndarray)

        lens_source_result.constant.lens_galaxies.lens = g.Galaxy(redshift=0.5)
        lens_source_result.constant.source_galaxies.source = g.Galaxy(redshift=1.0)

    def test_multi_plane_image_dict(self, multi_plane_result):
        image_dict = multi_plane_result.image_2d_dict
        assert isinstance(image_dict[("galaxies", "lens")], np.ndarray)
        assert isinstance(image_dict[("galaxies", "source")], np.ndarray)

        multi_plane_result.constant.galaxies.lens = g.Galaxy(redshift=0.5)

        image_dict = multi_plane_result.image_2d_dict
        assert (image_dict[("galaxies", "lens")] == np.zeros((5, 5))).all()
        assert isinstance(image_dict[("galaxies", "source")], np.ndarray)

    def test_galaxy_image_dict(self, lens_galaxy, source_galaxy, grid_stack_5x5,
                               convolver_image_5x5):
        tracer = rt.TracerImageSourcePlanes([lens_galaxy], [source_galaxy],
                                            grid_stack_5x5)

        assert len(tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=convolver_image_5x5)) == 2
        assert lens_galaxy in tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=convolver_image_5x5)
        assert source_galaxy in tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=convolver_image_5x5)

    def test__results_are_passed_to_new_analysis__sets_up_hyper_images(
            self,
            mask_function_5x5,
            results_collection_5x5,
            ccd_data_5x5
    ):
        # noinspection PyPep8Naming
        Phase = phase_imaging.LensSourcePlanePhase

        phase_5x5 = Phase(lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)),
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_5x5,
            phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5,
                                           results=results_collection_5x5)

        assert (analysis.hyper_model_image_1d == 5.0 * np.ones(9)).all()

        assert (analysis.hyper_galaxy_image_1d_path_dict[('g0',)] == 2.0 * np.ones(
            9)).all()
        assert (analysis.hyper_galaxy_image_1d_path_dict[('g1',)] == 3.0 * np.ones(
            9)).all()

    def test__results_are_passed_to_new_analysis__sets_up_hyper_cluster_images(self, mask_function_5x5, results_collection_5x5,
            ccd_data_5x5):

        # noinspection PyPep8Naming
        Phase = phase_imaging.LensSourcePlanePhase

        phase_5x5 = Phase(lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)),
                          optimizer_class=mock_pipeline.MockNLO, mask_function=mask_function_5x5,
                          cluster_pixel_scale=None, phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, results=results_collection_5x5)

        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g0',)] == 2.0 * np.ones(9)).all()
        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g1',)] == 3.0 * np.ones(9)).all()

        phase_5x5 = Phase(lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)),
                          optimizer_class=mock_pipeline.MockNLO, mask_function=mask_function_5x5,
                          cluster_pixel_scale=ccd_data_5x5.pixel_scale, phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, results=results_collection_5x5)

        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g0',)] == 2.0 * np.ones(9)).all()
        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g1',)] == 3.0 * np.ones(9)).all()
        assert len(analysis.hyper_galaxy_cluster_image_1d_path_dict[('g0',)]) == \
               analysis.lens_data.cluster.shape[0]
        assert len(analysis.hyper_galaxy_cluster_image_1d_path_dict[('g1',)]) == \
               analysis.lens_data.cluster.shape[0]

        phase_5x5 = Phase(lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)),
                          optimizer_class=mock_pipeline.MockNLO, mask_function=mask_function_5x5,
                          cluster_pixel_scale=ccd_data_5x5.pixel_scale*2.0, phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, results=results_collection_5x5)

        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g0',)] == 2.0 * np.ones(4)).all()
        assert (analysis.hyper_galaxy_cluster_image_1d_path_dict[('g1',)] == 3.0 * np.ones(4)).all()
        assert len(analysis.hyper_galaxy_cluster_image_1d_path_dict[('g0',)]) == \
               analysis.lens_data.cluster.shape[0]
        assert len(analysis.hyper_galaxy_cluster_image_1d_path_dict[('g1',)]) == \
               analysis.lens_data.cluster.shape[0]

    def test__image_in_results_has_masked_value_passsed__raises_error(self,
                                                                      mask_function_5x5,
                                                                      results_collection_5x5,
                                                                      ccd_data_5x5):
        # noinspection PyPep8Naming
        Phase = phase_imaging.LensSourcePlanePhase

        phase_5x5 = Phase(lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)),
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_5x5,
            phase_name='test_phase')

        results_collection_5x5[0].galaxy_images[0][2, 2] = 0.0

        with pytest.raises(exc.PhaseException):
            phase_5x5.make_analysis(data=ccd_data_5x5, results=results_collection_5x5)

    def test_associate_images_lens(self, lens_instance, lens_result, lens_data_5x5):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", lens_result)
        analysis = phase_imaging.LensPlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=None, positions_threshold=None,
            results=results_collection, uses_hyper_images=True)

        instance = analysis.associate_images(instance=lens_instance)

        hyper_model_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=lens_result.image_2d_dict[("lens_galaxies", "lens")])

        assert instance.lens_galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)
        assert instance.lens_galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)

    def test_associate_images_lens_source(self, lens_source_instance,
                                          lens_source_result, lens_data_5x5):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", lens_source_result)
        analysis = phase_imaging.LensSourcePlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=None, positions_threshold=None,
            results=results_collection, uses_hyper_images=True)

        instance = analysis.associate_images(lens_source_instance)

        hyper_lens_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=lens_source_result.image_2d_dict[("lens_galaxies", "lens")])
        hyper_source_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=lens_source_result.image_2d_dict[("source_galaxies", "source")])

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        assert instance.lens_galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)
        assert instance.source_galaxies.source.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)

        assert instance.lens_galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_lens_image_1d, 1.0e-4)
        assert instance.source_galaxies.source.hyper_galaxy_image_1d == pytest.approx(
            hyper_source_image_1d, 1.04e-4)

    def test_associate_images_multi_plane(self, multi_plane_instance,
                                          multi_plane_result, lens_data_5x5):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", multi_plane_result)
        analysis = phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=None, positions_threshold=None,
            results=results_collection, uses_hyper_images=True)

        instance = analysis.associate_images(instance=multi_plane_instance)

        hyper_lens_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "lens")])
        hyper_source_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "source")])

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        assert instance.galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_lens_image_1d, 1.0e-4)
        assert instance.galaxies.source.hyper_galaxy_image_1d == pytest.approx(
            hyper_source_image_1d, 1.0e-4)

        assert instance.galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)
        assert instance.galaxies.source.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4)

    def test_fit_uses_hyper_fit_correctly_multi_plane(self, multi_plane_instance,
                                                      multi_plane_result,
                                                      lens_data_5x5):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", multi_plane_result)
        analysis = phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_5x5, cosmology=cosmo.Planck15, positions_threshold=None,
            results=results_collection, uses_hyper_images=True)

        hyper_galaxy = g.HyperGalaxy(contribution_factor=1.0, noise_factor=1.0,
                                     noise_power=1.0)

        multi_plane_instance.galaxies.lens.hyper_galaxy = hyper_galaxy

        fit_figure_of_merit = analysis.fit(instance=multi_plane_instance)

        hyper_lens_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "lens")])
        hyper_source_image_1d = lens_data_5x5.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "source")])

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        g0 = g.Galaxy(redshift=0.5,
                      light_profile=multi_plane_instance.galaxies.lens.light,
                      mass_profile=multi_plane_instance.galaxies.lens.mass,
                      hyper_galaxy=hyper_galaxy,
                      hyper_model_image_1d=hyper_model_image_1d,
                      hyper_galaxy_image_1d=hyper_lens_image_1d,
                      hyper_minimum_value=0.0)
        g1 = g.Galaxy(redshift=1.0,
                      light_profile=multi_plane_instance.galaxies.source.light)

        tracer = rt.TracerImageSourcePlanes(
            lens_galaxies=[g0],
            source_galaxies=[g1],
            image_plane_grid_stack=lens_data_5x5.grid_stack
        )

        fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5,
                                                       tracer=tracer,
                                                       padded_tracer=None)

        assert (fit_figure_of_merit == fit.figure_of_merit).all()

    # def test__results_are_passed_to_new_analysis__associate_images_works(
    #         self, lens_source_instance, mask_function_5x5, results_collection_5x5,
    #         ccd_data_5x5):
    #     Phase = phase_imaging.LensSourcePlanePhase
    #
    #     phase_5x5 = Phase(optimizer_class=mock_pipeline.MockNLO,
    #                       mask_function=mask_function_5x5,
    #                       phase_name='test_phase')
    #
    #     analysis = phase_5x5.make_analysis(data=ccd_data_5x5,
    #                                        results=results_collection_5x5)
    #
    #     instance = analysis.associate_images(instance=lens_source_instance)
    #
    #     assert (analysis.hyper_model_image_1d == 5.0 * np.ones(9)).all()
    #
    #     print(instance.lens_galaxies.lens.hyper_galaxy_image_1d)
    #
    #     assert (instance.lens_galaxies.lens.hyper_galaxy_image_1d == 2.0 * np.ones(
    #         9)).all()
    #     assert (instance.source_galaxies.source.hyper_galaxy_image_1d == 3.0 * np.ones(
#         9)).all()



import os
from os import path

import numpy as np
import pytest
from astropy import cosmology as cosmo

import autofit as af
import autolens.pipeline.phase.phase_imaging
from autolens import exc
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.lens import lens_fit
from autolens.model.galaxy import galaxy as g, galaxy_model as gm
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.pipeline.phase import phase
from autolens.pipeline.phase import phase_imaging
from test.unit.mock.pipeline import mock_pipeline

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an array index, `arr[np.array(seq)]`, which will result "
    "either in an error or a different result.")

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(scope="session", autouse=True)
def do_something():
    af.conf.instance = af.conf.Config(
        '{}/../test_files/config/phase_5x5'.format(directory))


def clean_images():
    try:
        os.remove('{}/source_lens_phase/source_image_0.fits'.format(directory))
        os.remove('{}/source_lens_phase/lens_image_0.fits'.format(directory))
        os.remove('{}/source_lens_phase/model_image_0.fits'.format(directory))
    except FileNotFoundError:
        pass
    af.conf.instance.data_path = directory


class TestPhase(object):

    def test_set_constants(self, phase_5x5):
        phase_5x5.lens_galaxies = [g.Galaxy(redshift=0.5)]
        assert phase_5x5.optimizer.variable.lens_galaxies == [g.Galaxy(redshift=0.5)]

    def test_set_variables(self, phase_5x5):
        phase_5x5.lens_galaxies = [gm.GalaxyModel(redshift=0.5)]
        assert phase_5x5.optimizer.variable.lens_galaxies == [
            gm.GalaxyModel(redshift=0.5)]

    def test_make_analysis(self, phase_5x5, ccd_data_5x5, lens_data_5x5):
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)

        assert analysis.last_results is None
        assert analysis.lens_data.unmasked_image == ccd_data_5x5.image
        assert analysis.lens_data.unmasked_noise_map == ccd_data_5x5.noise_map
        assert analysis.lens_data.image_2d == lens_data_5x5.image_2d
        assert analysis.lens_data.noise_map_2d == lens_data_5x5.noise_map_2d

    def test_make_analysis__mask_input_uses_mask__no_mask_uses_mask_function(self,
                                                                             phase_5x5,
                                                                             ccd_data_5x5):
        # If an input mask is supplied and there is no mask function, we use mask input.

        phase_5x5.mask_function = None

        mask_input = msk.Mask.circular(shape=ccd_data_5x5.shape, pixel_scale=1,
                                       radius_arcsec=1.5)

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=mask_input)
        assert (analysis.lens_data.mask_2d == mask_input).all()

        # If a mask function is suppled, we should use this mask, regardless of whether an input mask is supplied.

        def mask_function(image):
            return msk.Mask.circular(shape=image.shape, pixel_scale=1,
                                     radius_arcsec=0.3)

        mask_from_function = mask_function(image=ccd_data_5x5.image)
        phase_5x5.mask_function = mask_function

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=None)
        assert (analysis.lens_data.mask_2d == mask_from_function).all()
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=mask_input)
        assert (analysis.lens_data.mask_2d == mask_from_function).all()

        # If no mask is suppled, nor a mask function, we should use the default mask. This extends behind the edge of
        # 5x5 image, so will raise a MaskException.

        phase_5x5.mask_function = None

        with pytest.raises(exc.MaskException):
            phase_5x5.make_analysis(data=ccd_data_5x5, mask=None)

    def test_make_analysis__mask_input_uses_mask__inner_mask_radius_included_which_masks_centre(
            self, phase_5x5, ccd_data_5x5):
        # If an input mask is supplied and there is no mask function, we use mask input.

        phase_5x5.mask_function = None
        phase_5x5.inner_mask_radii = 0.5

        mask_input = msk.Mask.circular(shape=ccd_data_5x5.shape, pixel_scale=1,
                                       radius_arcsec=1.5)

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=mask_input)

        # The inner circulaar mask radii of 0.5" masks only the central pixel of the mask

        mask_input[2, 2] = True

        assert (analysis.lens_data.mask_2d == mask_input).all()

        # If a mask function is supplied, we should use this mask, regardless of whether an input mask is supplied.

        def mask_function(image):
            return msk.Mask.circular(shape=image.shape, pixel_scale=1,
                                     radius_arcsec=1.4)

        mask_from_function = mask_function(image=ccd_data_5x5.image)

        # The inner circulaar mask radii of 1.0" masks the centra pixels of the mask
        mask_from_function[2, 2] = True

        phase_5x5.mask_function = mask_function

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=None)
        assert (analysis.lens_data.mask_2d == mask_from_function).all()
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5, mask=mask_input)
        assert (analysis.lens_data.mask_2d == mask_from_function).all()

        # If no mask is suppled, nor a mask function, we should use the default mask.

        phase_5x5.mask_function = None

        with pytest.raises(exc.MaskException):
            phase_5x5.make_analysis(data=ccd_data_5x5, mask=None)

    def test_make_analysis__positions_are_input__are_used_in_analysis(self, phase_5x5,
                                                                      ccd_data_5x5):
        # If position threshold is input (not None) and positions are input, make the positions part of the lens data.

        phase_5x5.positions_threshold = 0.2

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5,
                                           positions=[[[1.0, 1.0], [2.0, 2.0]]])
        assert (analysis.lens_data.positions[0][0] == np.array([1.0, 1.0])).all()
        assert (analysis.lens_data.positions[0][1] == np.array([2.0, 2.0])).all()

        # If position threshold is input (not None) and but no positions are supplied, raise an error

        with pytest.raises(exc.PhaseException):
            phase_5x5.make_analysis(data=ccd_data_5x5, positions=None)
            phase_5x5.make_analysis(data=ccd_data_5x5)

        # If positions threshold is None, positions should always be None.

        phase_5x5.positions_threshold = None
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5,
                                           positions=[[[1.0, 1.0], [2.0, 2.0]]])
        assert analysis.lens_data.positions is None

    def test_make_analysis__interp_pixel_scale_is_input__interp_grid_used_in_analysis(
            self, phase_5x5, ccd_data_5x5):
        # If use positions is true and positions are input, make the positions part of the lens data.

        phase_5x5.interp_pixel_scale = 0.1

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        assert analysis.lens_data.interp_pixel_scale == 0.1
        assert hasattr(analysis.lens_data.grid_stack.regular, 'interpolator')
        assert hasattr(analysis.lens_data.grid_stack.sub, 'interpolator')
        assert hasattr(analysis.lens_data.grid_stack.blurring, 'interpolator')
        assert hasattr(analysis.lens_data.padded_grid_stack.regular, 'interpolator')
        assert hasattr(analysis.lens_data.padded_grid_stack.sub, 'interpolator')

    def test__make_analysis__phase_info_is_made(self, phase_5x5, ccd_data_5x5):
        phase_5x5.make_analysis(data=ccd_data_5x5)

        file_phase_info = "{}/{}".format(phase_5x5.optimizer.phase_output_path,
                                         'phase.info')

        phase_info = open(file_phase_info, 'r')

        optimizer = phase_info.readline()
        sub_grid_size = phase_info.readline()
        image_psf_shape = phase_info.readline()
        pixelization_psf_shape = phase_info.readline()
        positions_threshold = phase_info.readline()
        cosmology = phase_info.readline()
        auto_link_priors = phase_info.readline()

        phase_info.close()

        assert optimizer == 'Optimizer = MockNLO \n'
        assert sub_grid_size == 'Sub-grid size = 2 \n'
        assert image_psf_shape == 'Image PSF shape = None \n'
        assert pixelization_psf_shape == 'Pixelization PSF shape = None \n'
        assert positions_threshold == 'Positions Threshold = None \n'
        assert cosmology == 'Cosmology = FlatLambdaCDM(name="Planck15", H0=67.7 km / (Mpc s), Om0=0.307, Tcmb0=2.725 K, ' \
                            'Neff=3.05, m_nu=[0.   0.   0.06] eV, Ob0=0.0486) \n'
        assert auto_link_priors == 'Auto Link Priors = False \n'

    def test_fit(self, ccd_data_5x5, mask_function_5x5):
        clean_images()

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            optimizer_class=mock_pipeline.MockNLO,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, light=lp.EllipticalSersic)),
            source_galaxies=dict(
                source=gm.GalaxyModel(redshift=0.5, light=lp.EllipticalSersic)),
            mask_function=mask_function_5x5,
            phase_name='test_phase_test_fit')

        result = phase_5x5.run(data=ccd_data_5x5)
        assert isinstance(result.constant.lens_galaxies[0], g.Galaxy)
        assert isinstance(result.constant.source_galaxies[0], g.Galaxy)

    def test_customize(self, mask_function_5x5, results_5x5, results_collection_5x5,
                       ccd_data_5x5):
        class MyPlanePhaseAnd(
            phase_imaging.LensSourcePlanePhase):

            def pass_priors(self, results):
                self.lens_galaxies = results.last.constant.lens_galaxies
                self.source_galaxies = results.last.variable.source_galaxies

        galaxy = g.Galaxy(redshift=0.5)
        galaxy_model = gm.GalaxyModel(redshift=0.5)

        setattr(results_5x5.constant, "lens_galaxies", [galaxy])
        setattr(results_5x5.variable, "source_galaxies", [galaxy_model])

        phase_5x5 = MyPlanePhaseAnd(optimizer_class=mock_pipeline.MockNLO,
                                    mask_function=mask_function_5x5,
                                    phase_name='test_phase')

        phase_5x5.make_analysis(data=ccd_data_5x5, results=results_collection_5x5)
        phase_5x5.pass_priors(results_collection_5x5)

        assert phase_5x5.lens_galaxies == [galaxy]
        assert phase_5x5.source_galaxies == [galaxy_model]

    def test_default_mask_function(self, phase_5x5, ccd_data_5x5):
        lens_data = ld.LensData(ccd_data=ccd_data_5x5,
                                mask=phase_5x5.mask_function(ccd_data_5x5.image))
        assert len(lens_data.image_1d) == 9

    def test_duplication(self):
        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)),
            source_galaxies=dict(source=gm.GalaxyModel(redshift=0.5)),
            phase_name='test_phase')

        phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase')

        assert phase_5x5.lens_galaxies is not None
        assert phase_5x5.source_galaxies is not None

    def test_modify_image(self, mask_function_5x5, ccd_data_5x5):
        class MyPhase(phase_imaging.PhaseImaging):

            def modify_image(self, image, results):
                assert ccd_data_5x5.image.shape == image.shape
                image = 20.0 * np.ones(shape=(5, 5))
                return image

        phase_5x5 = MyPhase(phase_name='phase_5x5', mask_function=mask_function_5x5)
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        assert (analysis.lens_data.unmasked_image == 20.0 * np.ones(shape=(5, 5))).all()
        assert (analysis.lens_data.image_1d == 20.0 * np.ones(shape=9)).all()

    def test__check_if_phase_uses_inversion(self, mask_function_5x5):
        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)))

        assert phase_5x5.uses_inversion is False

        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.Constant)))

        assert phase_5x5.uses_inversion is True

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)),
            source_galaxies=dict(source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_inversion is False

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.Constant)),
            source_galaxies=dict(source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_inversion is True

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)),
            source_galaxies=dict(
                source=gm.GalaxyModel(redshift=1.0, pixelization=pix.Rectangular,
                                      regularization=reg.Constant)))

        assert phase_5x5.uses_inversion is True

    def test__check_if_phase_uses_hyper_images(self, mask_function_5x5):
        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)))

        assert phase_5x5.uses_hyper_images is False

        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.AdaptiveBrightness)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.Constant)),
            source_galaxies=dict(source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_hyper_images is False

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.AdaptiveBrightness)),
            source_galaxies=dict(source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5)),
            source_galaxies=dict(
                source=gm.GalaxyModel(redshift=1.0, pixelization=pix.Rectangular,
                                      regularization=reg.AdaptiveBrightness)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5)),
            source_galaxies=dict(
                source=gm.GalaxyModel(redshift=1.0, hyper_galaxy=g.HyperGalaxy)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.MultiPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            galaxies=dict(lens=gm.GalaxyModel(redshift=0.5),
                          source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_hyper_images is False

        phase_5x5 = phase_imaging.MultiPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            galaxies=dict(lens=gm.GalaxyModel(redshift=0.5),
                          source=gm.GalaxyModel(redshift=1.0,
                                                hyper_galaxy=g.HyperGalaxy)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.MultiPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy),
                          source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.MultiPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            galaxies=dict(lens=gm.GalaxyModel(redshift=0.5),
                          source=gm.GalaxyModel(redshift=1.0, pixelization=pix.Rectangular,
                                    regularization=reg.AdaptiveBrightness)))

        assert phase_5x5.uses_hyper_images is True

        phase_5x5 = phase_imaging.MultiPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, pixelization=pix.Rectangular,
                                    regularization=reg.AdaptiveBrightness),
                source=gm.GalaxyModel(redshift=1.0)))

        assert phase_5x5.uses_hyper_images is True

    def test__adds_pixelization_grid_to_grid_stack_if_required(self, ccd_data_5x5,
                                                               mask_function_5x5):
        phase_5x5 = phase_imaging.MultiPlanePhase(phase_name='test_phase',
                                       mask_function=mask_function_5x5)

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)

        galaxy = g.Galaxy(redshift=0.5)

        grid_stack = analysis.add_grids_to_grid_stack(galaxies=[galaxy, galaxy],
                                                      grid_stack=analysis.lens_data.grid_stack)

        assert (grid_stack.pixelization == np.array([[0.0, 0.0]])).all()

        galaxy_pix_which_doesnt_use_pix_grid = g.Galaxy(redshift=0.5,
                                                        pixelization=pix.Rectangular(),
                                                        regularization=reg.Constant())

        grid_stack = analysis.add_grids_to_grid_stack(
            galaxies=[galaxy_pix_which_doesnt_use_pix_grid],
            grid_stack=analysis.lens_data.grid_stack)

        assert (grid_stack.pixelization == np.array([[0.0, 0.0]])).all()

        galaxy_pix_which_uses_pix_grid = g.Galaxy(redshift=0.5,
                                                  pixelization=pix.VoronoiMagnification(),
                                                  regularization=reg.Constant())

        grid_stack = analysis.add_grids_to_grid_stack(galaxies=[galaxy_pix_which_uses_pix_grid],
                                                      grid_stack=analysis.lens_data.grid_stack)

        assert (grid_stack.pixelization == np.array(
            [[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
             [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
             [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])).all()

        galaxy_pix_which_uses_brightness = g.Galaxy(redshift=0.5,
                                                    pixelization=pix.VoronoiBrightnessImage(
                                                        pixels=9),
                                                    regularization=reg.Constant())

        galaxy_pix_which_uses_brightness.hyper_galaxy_cluster_image_1d = np.array([0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0])

        grid_stack = analysis.add_grids_to_grid_stack(
            galaxies=[galaxy_pix_which_uses_brightness],
            grid_stack=analysis.lens_data.grid_stack)

        assert (grid_stack.pixelization == np.array(
            [[0.0, 1.0], [1.0, -1.0], [-1.0, -1.0],
             [-1.0, 1.0], [0.0, -1.0], [1.0, 1.0],
             [-1.0, 0.0], [0.0, 0.0], [1.0, 0.0]])).all()

    def test__phase_with_no_inversion__convolver_mapping_matrix_of_lens_data_is_none(
            self, ccd_data_5x5, mask_function_5x5):
        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name='test_phase', mask_function=mask_function_5x5,
            lens_galaxies=dict(lens=gm.GalaxyModel(redshift=0.5)))

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)

        assert analysis.lens_data.convolver_mapping_matrix is None

    def test__lens_data_is_binned_up(self, ccd_data_5x5, mask_5x5_1_pix,
                                     mask_function_5x5_1_pix):
        binned_up_ccd_data = ccd_data_5x5.new_ccd_data_with_binned_up_arrays(
            bin_up_factor=2)
        binned_up_mask = mask_5x5_1_pix.binned_up_mask_from_mask(bin_up_factor=2)

        phase_5x5 = phase_imaging.PhaseImaging(
            phase_name='phase_5x5', bin_up_factor=2,
            mask_function=mask_function_5x5_1_pix)
        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        assert (analysis.lens_data.unmasked_image == binned_up_ccd_data.image).all()
        assert (analysis.lens_data.psf == binned_up_ccd_data.psf).all()
        assert (
                analysis.lens_data.unmasked_noise_map == binned_up_ccd_data.noise_map).all()

        assert (analysis.lens_data.mask_2d == binned_up_mask).all()

        lens_data = ld.LensData(ccd_data=ccd_data_5x5, mask=mask_5x5_1_pix)
        binned_up_lens_data = lens_data.new_lens_data_with_binned_up_ccd_data_and_mask(
            bin_up_factor=2)

        assert (analysis.lens_data.image_2d == binned_up_lens_data.image_2d).all()
        assert (analysis.lens_data.psf == binned_up_lens_data.psf).all()
        assert (
                analysis.lens_data.noise_map_2d == binned_up_lens_data.noise_map_2d).all()

        assert (analysis.lens_data.mask_2d == binned_up_lens_data.mask_2d).all()

        assert (analysis.lens_data.image_1d == binned_up_lens_data.image_1d).all()
        assert (
                analysis.lens_data.noise_map_1d == binned_up_lens_data.noise_map_1d).all()

    def test__tracer_for_instance__includes_cosmology(self, ccd_data_5x5,
                                                      mask_function_5x5):
        lens_galaxy = g.Galaxy(redshift=0.5)
        source_galaxy = g.Galaxy(redshift=0.5)

        phase_5x5 = phase_imaging.LensPlanePhase(
            mask_function=mask_function_5x5,
            lens_galaxies=[lens_galaxy], cosmology=cosmo.FLRW, phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        instance = phase_5x5.variable.instance_from_unit_vector([])
        tracer = analysis.tracer_for_instance(instance=instance)
        padded_tracer = analysis.padded_tracer_for_instance(instance=instance)

        assert tracer.image_plane.galaxies[0] == lens_galaxy
        assert tracer.cosmology == cosmo.FLRW
        assert padded_tracer.image_plane.galaxies[0] == lens_galaxy
        assert padded_tracer.cosmology == cosmo.FLRW

        phase_5x5 = phase_imaging.LensSourcePlanePhase(
            mask_function=mask_function_5x5,
            lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
            cosmology=cosmo.FLRW,
            phase_name='test_phase')

        analysis = phase_5x5.make_analysis(ccd_data_5x5)
        instance = phase_5x5.variable.instance_from_unit_vector([])
        tracer = analysis.tracer_for_instance(instance)
        padded_tracer = analysis.padded_tracer_for_instance(instance)

        assert tracer.image_plane.galaxies[0] == lens_galaxy
        assert tracer.source_plane.galaxies[0] == source_galaxy
        assert tracer.cosmology == cosmo.FLRW
        assert padded_tracer.image_plane.galaxies[0] == lens_galaxy
        assert padded_tracer.source_plane.galaxies[0] == source_galaxy
        assert padded_tracer.cosmology == cosmo.FLRW

        galaxy_0 = g.Galaxy(redshift=0.1)
        galaxy_1 = g.Galaxy(redshift=0.2)
        galaxy_2 = g.Galaxy(redshift=0.3)

        phase_5x5 = phase_imaging.MultiPlanePhase(
            mask_function=mask_function_5x5,
            galaxies=[galaxy_0, galaxy_1, galaxy_2], cosmology=cosmo.WMAP7,
            phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        instance = phase_5x5.variable.instance_from_unit_vector([])
        tracer = analysis.tracer_for_instance(instance)
        padded_tracer = analysis.padded_tracer_for_instance(instance)

        assert tracer.planes[0].galaxies[0] == galaxy_0
        assert tracer.planes[1].galaxies[0] == galaxy_1
        assert tracer.planes[2].galaxies[0] == galaxy_2
        assert tracer.cosmology == cosmo.WMAP7
        assert padded_tracer.planes[0].galaxies[0] == galaxy_0
        assert padded_tracer.planes[1].galaxies[0] == galaxy_1
        assert padded_tracer.planes[2].galaxies[0] == galaxy_2
        assert padded_tracer.cosmology == cosmo.WMAP7

    def test__fit_figure_of_merit__matches_correct_fit_given_galaxy_profiles(
            self,
            ccd_data_5x5,
            mask_function_5x5
    ):
        # noinspection PyTypeChecker
        lens_galaxy = g.Galaxy(redshift=0.5, light=lp.EllipticalSersic(intensity=0.1))

        phase_5x5 = phase_imaging.LensPlanePhase(
            lens_galaxies=[lens_galaxy], mask_function=mask_function_5x5,
            cosmology=cosmo.FLRW, phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        instance = phase_5x5.variable.instance_from_unit_vector([])

        fit_figure_of_merit = analysis.fit(instance=instance)

        mask = phase_5x5.mask_function(image=ccd_data_5x5.image)
        lens_data = ld.LensData(ccd_data=ccd_data_5x5, mask=mask)
        tracer = analysis.tracer_for_instance(instance=instance)
        fit = lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)

        assert fit.likelihood == fit_figure_of_merit

    def test__phase_can_receive_list_of_galaxy_models(self):
        phase_5x5 = phase_imaging.LensPlanePhase(
            lens_galaxies=dict(lens=gm.GalaxyModel(
                sersic=lp.EllipticalSersic,
                sis=mp.SphericalIsothermal,
                redshift=g.Redshift),
                lens1=gm.GalaxyModel(sis=mp.SphericalIsothermal,
                                     redshift=g.Redshift)),
            optimizer_class=af.MultiNest, phase_name='test_phase')

        instance = phase_5x5.optimizer.variable.instance_from_physical_vector(
            [0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.8, 0.1, 0.2, 0.3,
             0.4, 0.9, 0.5, 0.7, 0.8])

        assert instance.lens_galaxies[0].sersic.centre[0] == 0.2
        assert instance.lens_galaxies[0].sis.centre[0] == 0.1
        assert instance.lens_galaxies[0].sis.centre[1] == 0.2
        assert instance.lens_galaxies[0].sis.einstein_radius == 0.3
        assert instance.lens_galaxies[0].redshift == 0.4
        assert instance.lens_galaxies[1].sis.centre[0] == 0.9
        assert instance.lens_galaxies[1].sis.centre[1] == 0.5
        assert instance.lens_galaxies[1].sis.einstein_radius == 0.7
        assert instance.lens_galaxies[1].redshift == 0.8

        class LensPlanePhase2(phase_imaging.LensPlanePhase):
            # noinspection PyUnusedLocal
            def pass_models(self, results):
                self.lens_galaxies[0].sis.einstein_radius = 10.0

        phase_5x5 = LensPlanePhase2(
            lens_galaxies=dict(lens=gm.GalaxyModel(sersic=lp.EllipticalSersic,
                                                   sis=mp.SphericalIsothermal,
                                                   redshift=g.Redshift),
                               lens1=gm.GalaxyModel(sis=mp.SphericalIsothermal,
                                                    redshift=g.Redshift)),
            optimizer_class=af.MultiNest, phase_name='test_phase')

        # noinspection PyTypeChecker
        phase_5x5.pass_models(None)

        instance = phase_5x5.optimizer.variable.instance_from_physical_vector(
            [0.01, 0.02, 0.23, 0.04, 0.05, 0.06, 0.87, 0.1, 0.2,
             0.4, 0.5, 0.5, 0.7, 0.8])

        assert instance.lens_galaxies[0].sersic.centre[0] == 0.01
        assert instance.lens_galaxies[0].sis.centre[0] == 0.1
        assert instance.lens_galaxies[0].sis.centre[1] == 0.2
        assert instance.lens_galaxies[0].sis.einstein_radius == 10.0
        assert instance.lens_galaxies[0].redshift == 0.4
        assert instance.lens_galaxies[1].sis.centre[0] == 0.5
        assert instance.lens_galaxies[1].sis.centre[1] == 0.5
        assert instance.lens_galaxies[1].sis.einstein_radius == 0.7
        assert instance.lens_galaxies[1].redshift == 0.8


class TestResult(object):

    def test__results_of_phase_are_available_as_properties(self, ccd_data_5x5,
                                                           mask_function_5x5):
        clean_images()

        phase_5x5 = phase_imaging.LensPlanePhase(
            optimizer_class=mock_pipeline.MockNLO, mask_function=mask_function_5x5,
            lens_galaxies=[
                g.Galaxy(redshift=0.5, light=lp.EllipticalSersic(intensity=1.0))],
            phase_name='test_phase_2')

        result = phase_5x5.run(data=ccd_data_5x5)

        assert isinstance(result, phase.AbstractPhase.Result)

    def test__fit_figure_of_merit__matches_correct_fit_given_galaxy_profiles(
            self,
            ccd_data_5x5,
            mask_function_5x5):
        lens_galaxy = g.Galaxy(redshift=0.5, light=lp.EllipticalSersic(intensity=0.1))

        phase_5x5 = phase_imaging.LensPlanePhase(
            mask_function=mask_function_5x5, lens_galaxies=[lens_galaxy],
            cosmology=cosmo.FLRW,
            phase_name='test_phase')

        analysis = phase_5x5.make_analysis(data=ccd_data_5x5)
        instance = phase_5x5.variable.instance_from_unit_vector([])
        fit_figure_of_merit = analysis.fit(instance=instance)

        mask = phase_5x5.mask_function(image=ccd_data_5x5.image)
        lens_data = ld.LensData(ccd_data=ccd_data_5x5, mask=mask)
        tracer = analysis.tracer_for_instance(instance=instance)
        fit = lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)

        assert fit.likelihood == fit_figure_of_merit


class TestPhasePickle(object):

    # noinspection PyTypeChecker
    def test_assertion_failure(self, ccd_data_5x5, mask_function_5x5):
        def make_analysis(*args, **kwargs):
            return mock_pipeline.MockAnalysis(1, 1)

        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name="phase_name", mask_function=mask_function_5x5,
            optimizer_class=mock_pipeline.MockNLO,
            lens_galaxies=dict(
                lens=g.Galaxy(light=lp.EllipticalLightProfile, redshift=1))
        )

        phase_5x5.make_analysis = make_analysis
        result = phase_5x5.run(data=ccd_data_5x5, results=None, mask=None,
                               positions=None)
        assert result is not None

        phase_5x5 = phase_imaging.LensPlanePhase(
            phase_name="phase_name", mask_function=mask_function_5x5,
            optimizer_class=mock_pipeline.MockNLO,
            lens_galaxies=dict(
                lens=g.Galaxy(light=lp.EllipticalLightProfile, redshift=1))
        )

        phase_5x5.make_analysis = make_analysis
        result = phase_5x5.run(data=ccd_data_5x5, results=None, mask=None,
                               positions=None)
        assert result is not None

        class CustomPhase(phase_imaging.LensPlanePhase):
            def pass_priors(self, results):

                self.lens_galaxies.lens.light = lp.EllipticalLightProfile()

        phase_5x5 = CustomPhase(
            phase_name="phase_name", mask_function=mask_function_5x5,
            optimizer_class=mock_pipeline.MockNLO,
            lens_galaxies=dict(
                lens=g.Galaxy(light=lp.EllipticalLightProfile, redshift=1))
        )
        phase_5x5.make_analysis = make_analysis

        with pytest.raises(af.exc.PipelineException):
            phase_5x5.run(data=ccd_data_5x5, results=None, mask=None, positions=None)
